//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Device/Instrument/CoordSystem2D.cpp
//! @brief     Implements ICoordSystem classes.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Device/Instrument/CoordSystem2D.h"
#include "Base/Const/Units.h"
#include "Base/Math/Constants.h"
#include "Base/Pixel/RectangularPixel.h"
#include "Device/Beam/Beam.h"
#include "Device/Coord/AxisNames.h"
#include "Device/Detector/RectangularDetector.h"
#include "Device/Detector/SphericalDetector.h"
#include "Device/ProDetector/RegionOfInterest.h"
#include <algorithm>
#include <cmath>
#include <stdexcept>

namespace {

double getQ(double wavelength, double angle)
{
    return 4.0 * M_PI * std::sin(angle) / wavelength;
}

double axisAngle(size_t i_axis, kvector_t k_f)
{
    if (i_axis == 0)
        return k_f.phi();
    if (i_axis == 1)
        return M_PI_2 - k_f.theta();
    ASSERT(0);
}

} // namespace


//  ************************************************************************************************
//  class CoordSystem2D
//  ************************************************************************************************

CoordSystem2D::CoordSystem2D(const Beam& beam)
    : m_wavelength(beam.wavelength())
    , m_alpha_i(-beam.direction().alpha())
    , m_phi_i(beam.direction().phi())
{
}

CoordSystem2D::CoordSystem2D(const CoordSystem2D& other)
    : m_axes(other.m_axes)
    , m_wavelength(other.m_wavelength)
    , m_alpha_i(other.m_alpha_i)
    , m_phi_i(other.m_phi_i)
{
}

void CoordSystem2D::addAxisData(std::string name, double min, double max,
                                Axes::Coords default_units, size_t nbins)
{
    m_axes.emplace_back(AxisData{name, min, max, default_units, nbins});
}

double CoordSystem2D::calculateMin(size_t i_axis, Axes::Coords units) const
{
    ASSERT(i_axis < dimension());
    units = substituteDefaultUnits(units);
    const auto& axis_data = m_axes[i_axis];
    if (units == Axes::Coords::NBINS)
        return 0.0;
    return calculateValue(i_axis, units, axis_data.min);
}

double CoordSystem2D::calculateMax(size_t i_axis, Axes::Coords units) const
{
    ASSERT(i_axis < dimension());
    units = substituteDefaultUnits(units);
    const auto& axis_data = m_axes[i_axis];
    if (units == Axes::Coords::NBINS)
        return static_cast<double>(axis_data.nbins);
    return calculateValue(i_axis, units, axis_data.max);
}

size_t CoordSystem2D::axisSize(size_t i_axis) const
{
    ASSERT(i_axis < dimension());
    return m_axes[i_axis].nbins;
}

std::vector<Axes::Coords> CoordSystem2D::availableUnits() const
{
    return {Axes::Coords::NBINS, Axes::Coords::RADIANS, Axes::Coords::DEGREES};
}

std::unique_ptr<const IAxis> CoordSystem2D::createConvertedAxis(size_t i_axis,
                                                                Axes::Coords units) const
{
    const double min = calculateMin(i_axis, units);
    const double max = calculateMax(i_axis, units);
    const auto& axis_name = axisName(i_axis, units);
    const auto axis_size = axisSize(i_axis);
    return std::make_unique<FixedBinAxis>(axis_name, axis_size, min, max);
}

void CoordSystem2D::addDetectorAxis(const IDetector& detector, size_t i_axis)
{
    const auto& axis = detector.axis(i_axis);
    const auto* roi = detector.regionOfInterest();
    const auto& axis_name = axisName(i_axis);
    if (!roi) {
        addAxisData(axis_name, axis.lowerBound(), axis.upperBound(), defaultUnits(), axis.size());
        return;
    }
    auto roi_axis = roi->clipAxisToRoi(i_axis, axis);
    addAxisData(axis_name, roi_axis->lowerBound(), roi_axis->upperBound(), defaultUnits(),
                roi_axis->size());
}


//  ************************************************************************************************
//  class SphericalCoords
//  ************************************************************************************************

SphericalCoords::SphericalCoords(const SphericalDetector& detector, const Beam& beam)
    : CoordSystem2D(beam)
{
    ASSERT(detector.dimension() == 2);
    addDetectorAxis(detector, 0);
    addDetectorAxis(detector, 1);
}

SphericalCoords::SphericalCoords(const SphericalCoords& other) : CoordSystem2D(other) {}

SphericalCoords::~SphericalCoords() = default;

SphericalCoords* SphericalCoords::clone() const
{
    return new SphericalCoords(*this);
}

std::vector<Axes::Coords> SphericalCoords::availableUnits() const
{
    auto result = CoordSystem2D::availableUnits();
    result.push_back(Axes::Coords::QSPACE);
    return result;
}

double SphericalCoords::calculateValue(size_t i_axis, Axes::Coords units, double value) const
{
    switch (units) {
    case Axes::Coords::RADIANS:
        return value;
    case Axes::Coords::DEGREES:
        return Units::rad2deg(value);
    case Axes::Coords::QSPACE: {
        const auto k_i = vecOfLambdaAlphaPhi(m_wavelength, m_alpha_i, m_phi_i);
        if (i_axis == 0) {
            const auto k_f = vecOfLambdaAlphaPhi(m_wavelength, 0.0, value);
            return (k_i - k_f).y();
        } else if (i_axis == 1) {
            const auto k_f = vecOfLambdaAlphaPhi(m_wavelength, value, 0.0);
            return (k_f - k_i).z();
        }
        throw std::runtime_error("Error in SphericalCoords::calculateValue: "
                                 "incorrect axis index: "
                                 + std::to_string(static_cast<int>(i_axis)));
    }
    case Axes::Coords::QXQY: {
        const auto k_i = vecOfLambdaAlphaPhi(m_wavelength, m_alpha_i, m_phi_i);
        if (i_axis == 0) {
            const auto k_f = vecOfLambdaAlphaPhi(m_wavelength, 0.0, value);
            return (k_i - k_f).y();
        } else if (i_axis == 1) {
            const auto k_f = vecOfLambdaAlphaPhi(m_wavelength, value, 0.0);
            return (k_f - k_i).x();
        }
        throw std::runtime_error("Error in SphericalCoords::calculateValue: "
                                 "incorrect axis index: "
                                 + std::to_string(static_cast<int>(i_axis)));
    }
    default:
        throwUnitsError("SphericalCoords::calculateValue", availableUnits());
    }
}

std::vector<std::map<Axes::Coords, std::string>> SphericalCoords::createNameMaps() const
{
    return {AxisNames::sphericalAxis0, AxisNames::sphericalAxis1};
}


//  ************************************************************************************************
//  class ImageCoords
//  ************************************************************************************************

ImageCoords::ImageCoords(const RectangularDetector& detector, const Beam& beam)
    : CoordSystem2D(beam)
{
    ASSERT(detector.dimension() == 2);
    addDetectorAxis(detector, 0);
    addDetectorAxis(detector, 1);
    m_detector_pixel.reset(detector.regionOfInterestPixel());
}

ImageCoords::ImageCoords(const ImageCoords& other)
    : CoordSystem2D(other), m_detector_pixel(other.m_detector_pixel->clone())
{
}

ImageCoords::~ImageCoords() = default;

ImageCoords* ImageCoords::clone() const
{
    return new ImageCoords(*this);
}

std::vector<Axes::Coords> ImageCoords::availableUnits() const
{
    auto result = CoordSystem2D::availableUnits();
    result.push_back(Axes::Coords::QSPACE);
    result.push_back(Axes::Coords::MM);
    return result;
}

double ImageCoords::calculateValue(size_t i_axis, Axes::Coords units, double value) const
{
    ASSERT(m_wavelength > 0);
    if (units == Axes::Coords::MM)
        return value;
    const auto k00 = m_detector_pixel->getPosition(0.0, 0.0);
    const auto k01 = m_detector_pixel->getPosition(0.0, 1.0);
    const auto k10 = m_detector_pixel->getPosition(1.0, 0.0);
    const auto& max_pos = i_axis == 0 ? k10 : k01; // position of max along given axis
    const double shift = value - m_axes[i_axis].min;
    const kvector_t out_dir = k00 + shift * (max_pos - k00).unit();
    const kvector_t k_f = out_dir.unit() * M_TWOPI / m_wavelength;

    switch (units) {
    case Axes::Coords::RADIANS:
        return axisAngle(i_axis, k_f);
    case Axes::Coords::DEGREES:
        return Units::rad2deg(axisAngle(i_axis, k_f));
    case Axes::Coords::QSPACE: {
        const auto k_i = vecOfLambdaAlphaPhi(m_wavelength, m_alpha_i, m_phi_i);
        if (i_axis == 0)
            return (k_i - k_f).y();
        if (i_axis == 1)
            return (k_f - k_i).z();
    }
    case Axes::Coords::QXQY: {
        const auto k_i = vecOfLambdaAlphaPhi(m_wavelength, m_alpha_i, m_phi_i);
        if (i_axis == 0)
            return (k_i - k_f).y();
        if (i_axis == 1)
            return (k_f - k_i).x();
    }
    default:
        throwUnitsError("ImageCoords::calculateValue", availableUnits());
    }
    ASSERT(0);
}

std::vector<std::map<Axes::Coords, std::string>> ImageCoords::createNameMaps() const
{
    return {AxisNames::rectangularAxis0, AxisNames::rectangularAxis1};
}


//  ************************************************************************************************
//  class OffSpecularCoordinates
//  ************************************************************************************************

OffSpecularCoordinates::OffSpecularCoordinates(const IDetector2D& detector, const Beam& beam,
                                               const IAxis& alpha_axis)
    : CoordSystem2D(beam)
{
    ASSERT(detector.dimension() == 2);
    addAxisData(axisName(0), alpha_axis.lowerBound(), alpha_axis.upperBound(), defaultUnits(),
                alpha_axis.size());
    addDetectorYAxis(detector);
}

OffSpecularCoordinates::OffSpecularCoordinates(const OffSpecularCoordinates& other)
    : CoordSystem2D(other)
{
}

OffSpecularCoordinates::~OffSpecularCoordinates() = default;

OffSpecularCoordinates* OffSpecularCoordinates::clone() const
{
    return new OffSpecularCoordinates(*this);
}

Axes::Coords OffSpecularCoordinates::defaultUnits() const
{
    return Axes::Coords::DEGREES;
}

double OffSpecularCoordinates::calculateValue(size_t, Axes::Coords units, double value) const
{
    switch (units) {
    case Axes::Coords::RADIANS:
        return value;
    case Axes::Coords::DEGREES:
        return Units::rad2deg(value);
    default:
        throwUnitsError("OffSpecularCoordinates::calculateValue", availableUnits());
    }
}

std::vector<std::map<Axes::Coords, std::string>> OffSpecularCoordinates::createNameMaps() const
{
    return {AxisNames::offSpecularAxis0, AxisNames::offSpecularAxis1};
}

void OffSpecularCoordinates::addDetectorYAxis(const IDetector2D& detector)
{
    const auto& axis = detector.axis(1);
    const auto* roi = detector.regionOfInterest();
    const auto& axis_name = axisName(1);
    std::unique_ptr<IAxis> new_axis;
    if (roi)
        new_axis = roi->clipAxisToRoi(1, axis);
    else
        new_axis.reset(axis.clone());
    ASSERT(new_axis);

    if (const auto* rect_det = dynamic_cast<const RectangularDetector*>(&detector)) {
        std::unique_ptr<RectangularPixel> det_pixel(rect_det->regionOfInterestPixel());
        const auto k00 = det_pixel->getPosition(0.0, 0.0);
        const auto k01 = det_pixel->getPosition(0.0, 1.0);
        const double alpha_f_min = M_PI_2 - k00.theta();
        const double alpha_f_max = M_PI_2 - k01.theta();
        addAxisData(axis_name, alpha_f_min, alpha_f_max, defaultUnits(), new_axis->size());
    } else if (dynamic_cast<const SphericalDetector*>(&detector)) {
        const double alpha_f_min = new_axis->lowerBound();
        const double alpha_f_max = new_axis->upperBound();
        addAxisData(axis_name, alpha_f_min, alpha_f_max, defaultUnits(), new_axis->size());
    } else
        ASSERT(0);
}


//  ************************************************************************************************
//  class DepthProbeCoordinates
//  ************************************************************************************************

const std::string z_axis_name = "Position [nm]";

DepthProbeCoordinates::DepthProbeCoordinates(const Beam& beam, const IAxis& alpha_axis,
                                             const IAxis& z_axis)
    : CoordSystem2D(beam)
{
    const auto& alpha_axis_name = axisName(0);
    const auto& z_axis_name = axisName(1);
    addAxisData(alpha_axis_name, alpha_axis.lowerBound(), alpha_axis.upperBound(), defaultUnits(),
                alpha_axis.size());
    addAxisData(z_axis_name, z_axis.lowerBound(), z_axis.upperBound(), defaultUnits(),
                z_axis.size());
}

DepthProbeCoordinates::DepthProbeCoordinates(const DepthProbeCoordinates& other)
    : CoordSystem2D(other)
{
}

DepthProbeCoordinates::~DepthProbeCoordinates() = default;

DepthProbeCoordinates* DepthProbeCoordinates::clone() const
{
    return new DepthProbeCoordinates(*this);
}

std::vector<Axes::Coords> DepthProbeCoordinates::availableUnits() const
{
    auto result = CoordSystem2D::availableUnits();
    result.push_back(Axes::Coords::QSPACE);
    return result;
}

double DepthProbeCoordinates::calculateValue(size_t i_axis, Axes::Coords units, double value) const
{
    const auto& available_units = availableUnits();
    if (std::find(available_units.begin(), available_units.end(), units) == available_units.cend())
        throwUnitsError("DepthProbeCoordinates::checkUnits", available_units);

    if (i_axis == 1)
        return value; // unit conversions are not applied to sample position axis
    switch (units) {
    case Axes::Coords::DEGREES:
        return Units::rad2deg(value);
    case Axes::Coords::QSPACE:
        return getQ(m_wavelength, value);
    default:
        return value;
    }
}

std::vector<std::map<Axes::Coords, std::string>> DepthProbeCoordinates::createNameMaps() const
{
    return {AxisNames::specAxis, AxisNames::sampleDepthAxis};
}
