//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/coregui/Models/MaterialItem.cpp
//! @brief     Implements class MaterialItem
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/coregui/Models/MaterialDataItems.h"
#include "GUI/coregui/Models/SessionItemUtils.h"
#include "GUI/coregui/Models/VectorItem.h"
#include "GUI/coregui/Views/MaterialEditor/MaterialItemUtils.h"
#include "GUI/coregui/utils/GUIHelpers.h"
#include "Sample/Material/MaterialFactoryFuncs.h"

namespace {
const QString magnetization_tooltip = "Magnetization (A/m)";
}

const QString MaterialItem::P_COLOR = "Color";
const QString MaterialItem::P_MATERIAL_DATA = "Material data";
const QString MaterialItem::P_MAGNETIZATION = "Magnetization";
const QString MaterialItem::P_IDENTIFIER = "Identifier";

const QString MaterialItem::M_TYPE = "Material";

MaterialItem::MaterialItem() : SessionItem(M_TYPE)
{
    setItemName("Material");

    ExternalProperty color = MaterialItemUtils::colorProperty(QColor(Qt::red));
    addProperty(P_COLOR, color.variant())->setEditorType("ExtColorEditor");

    addGroupProperty(P_MATERIAL_DATA, "Material data group");
    addProperty<VectorItem>(P_MAGNETIZATION)->setToolTip(magnetization_tooltip);
    addProperty(P_IDENTIFIER, GUIHelpers::createUuid());
    getItem(P_IDENTIFIER)->setVisible(false);
}

//! Turns material into refractive index material.

void MaterialItem::setRefractiveIndex(const double delta, const double beta)
{
    auto refractiveData = setGroupProperty(P_MATERIAL_DATA, MaterialRefractiveDataItem::M_TYPE);
    refractiveData->setItemValue(MaterialRefractiveDataItem::P_DELTA, delta);
    refractiveData->setItemValue(MaterialRefractiveDataItem::P_BETA, beta);
}

//! Turns material into SLD based material.

void MaterialItem::setScatteringLengthDensity(const complex_t sld)
{
    auto sldData = setGroupProperty(P_MATERIAL_DATA, MaterialSLDDataItem::M_TYPE);
    sldData->setItemValue(MaterialSLDDataItem::P_SLD_REAL, sld.real());
    sldData->setItemValue(MaterialSLDDataItem::P_SLD_IMAG, sld.imag());
}

bool MaterialItem::hasRefractiveIndex() const
{
    SessionItem* data = getGroupItem(P_MATERIAL_DATA);
    ASSERT(data);
    return dynamic_cast<MaterialRefractiveDataItem*>(data) != nullptr;
}

double MaterialItem::refractiveIndexDelta() const
{
    ASSERT(hasRefractiveIndex());
    return getGroupItem(P_MATERIAL_DATA)
        ->getItemValue(MaterialRefractiveDataItem::P_DELTA).toDouble();
}

double MaterialItem::refractiveIndexBeta() const
{
    ASSERT(hasRefractiveIndex());
    return getGroupItem(P_MATERIAL_DATA)
        ->getItemValue(MaterialRefractiveDataItem::P_BETA).toDouble();
}

complex_t MaterialItem::scatteringLengthDensity() const
{
    ASSERT(!hasRefractiveIndex());
    SessionItem* ri = getGroupItem(P_MATERIAL_DATA);
    double real = ri->getItemValue(MaterialSLDDataItem::P_SLD_REAL).toDouble();
    double imag = ri->getItemValue(MaterialSLDDataItem::P_SLD_IMAG).toDouble();
    return complex_t(real, imag);
}

QString MaterialItem::identifier() const
{
    return getItemValue(P_IDENTIFIER).toString();
}

void MaterialItem::setIdentifier(const QString& id)
{
    setItemValue(P_IDENTIFIER, id);
}

QColor MaterialItem::color() const
{
    ExternalProperty property = getItemValue(P_COLOR).value<ExternalProperty>();
    return property.color();
}

void MaterialItem::setColor(const QColor& color)
{
    setItemValue(P_COLOR, MaterialItemUtils::colorProperty(color).variant());
}

void MaterialItem::setMagnetization(const kvector_t& magnetization)
{
    item<VectorItem>(P_MAGNETIZATION)->setVector(magnetization);
}

void MaterialItem::hideMagnetization()
{
    getItem(MaterialItem::P_MAGNETIZATION)->setVisible(false);
}

std::unique_ptr<Material> MaterialItem::createMaterial() const
{
    auto dataItem = getGroupItem(P_MATERIAL_DATA);
    auto magnetization = item<VectorItem>(P_MAGNETIZATION)->getVector();
    auto name = itemName().toStdString();

    if (dataItem->modelType() == MaterialRefractiveDataItem::M_TYPE) {
        double delta = dataItem->getItemValue(MaterialRefractiveDataItem::P_DELTA).toDouble();
        double beta = dataItem->getItemValue(MaterialRefractiveDataItem::P_BETA).toDouble();
        return std::make_unique<Material>(HomogeneousMaterial(name, delta, beta, magnetization));

    } else if (dataItem->modelType() == MaterialSLDDataItem::M_TYPE) {
        double sld_real = dataItem->getItemValue(MaterialSLDDataItem::P_SLD_REAL).toDouble();
        double sld_imag = dataItem->getItemValue(MaterialSLDDataItem::P_SLD_IMAG).toDouble();
        return std::make_unique<Material>(MaterialBySLD(name, sld_real, sld_imag, magnetization));
    }

    throw GUIHelpers::Error("MaterialItem::createMaterial() -> Error. "
                            "Not implemented material type");
}
