#include "Base/Const/Units.h"
#include "Device/Beam/Beam.h"
#include "Device/Detector/SphericalDetector.h"
#include "Device/Instrument/CoordSystem2D.h"
#include "Tests/GTestWrapper/google_test.h"

class OffSpecularCoordinatesTest : public ::testing::Test {
public:
    OffSpecularCoordinatesTest();

protected:
    SphericalDetector m_detector;
    FixedBinAxis m_alpha_i_axis;
    Beam m_beam;
};

OffSpecularCoordinatesTest::OffSpecularCoordinatesTest()
    : m_detector(100, 0.0, 5.0 * Units::deg, 70, -2.0 * Units::deg, 1.5)
    , m_alpha_i_axis("alpha_i", 51, 0.0, 7.0 * Units::deg)
    , m_beam(1.0, 1.0, {1.0 * Units::deg, 0.0})
{
}

TEST_F(OffSpecularCoordinatesTest, OffSpecularCoordinates)
{
    OffSpecularCoordinates converter(m_detector, m_beam, m_alpha_i_axis);

    EXPECT_EQ(converter.dimension(), 2u);

    EXPECT_DOUBLE_EQ(converter.calculateMin(0, Axes::Coords::UNDEFINED), 0.0);
    EXPECT_DOUBLE_EQ(converter.calculateMin(0, Axes::Coords::NBINS), 0.0);
    EXPECT_DOUBLE_EQ(converter.calculateMin(0, Axes::Coords::RADIANS), 0.0);
    EXPECT_DOUBLE_EQ(converter.calculateMin(0, Axes::Coords::DEGREES), 0.0);
    EXPECT_THROW(converter.calculateMin(0, Axes::Coords::QSPACE), std::runtime_error);
    EXPECT_THROW(converter.calculateMin(0, Axes::Coords::MM), std::runtime_error);

    EXPECT_DOUBLE_EQ(converter.calculateMax(0, Axes::Coords::UNDEFINED), 7.0);
    EXPECT_DOUBLE_EQ(converter.calculateMax(0, Axes::Coords::NBINS), 51.0);
    EXPECT_DOUBLE_EQ(converter.calculateMax(0, Axes::Coords::RADIANS), Units::deg2rad(7.0));
    EXPECT_DOUBLE_EQ(converter.calculateMax(0, Axes::Coords::DEGREES), 7.0);
    EXPECT_THROW(converter.calculateMax(0, Axes::Coords::QSPACE), std::runtime_error);
    EXPECT_THROW(converter.calculateMax(0, Axes::Coords::MM), std::runtime_error);

    EXPECT_DOUBLE_EQ(converter.calculateMin(1, Axes::Coords::UNDEFINED), -2.0);
    EXPECT_DOUBLE_EQ(converter.calculateMin(1, Axes::Coords::NBINS), 0.0);
    EXPECT_DOUBLE_EQ(converter.calculateMin(1, Axes::Coords::RADIANS), Units::deg2rad(-2.0));
    EXPECT_DOUBLE_EQ(converter.calculateMin(1, Axes::Coords::DEGREES), -2.0);
    EXPECT_THROW(converter.calculateMin(1, Axes::Coords::QSPACE), std::runtime_error);
    EXPECT_THROW(converter.calculateMin(1, Axes::Coords::MM), std::runtime_error);

    EXPECT_DOUBLE_EQ(converter.calculateMax(1, Axes::Coords::UNDEFINED), Units::rad2deg(1.5));
    EXPECT_DOUBLE_EQ(converter.calculateMax(1, Axes::Coords::NBINS), 70.0);
    EXPECT_DOUBLE_EQ(converter.calculateMax(1, Axes::Coords::RADIANS), 1.5);
    EXPECT_DOUBLE_EQ(converter.calculateMax(1, Axes::Coords::DEGREES), Units::rad2deg(1.5));
    EXPECT_THROW(converter.calculateMax(1, Axes::Coords::QSPACE), std::runtime_error);
    EXPECT_THROW(converter.calculateMax(1, Axes::Coords::MM), std::runtime_error);

    EXPECT_THROW(converter.calculateMin(2, Axes::Coords::UNDEFINED), std::runtime_error);
    EXPECT_THROW(converter.calculateMax(2, Axes::Coords::UNDEFINED), std::runtime_error);

    auto axis = converter.createConvertedAxis(0, Axes::Coords::UNDEFINED);
    EXPECT_TRUE(dynamic_cast<const FixedBinAxis*>(axis.get()));
    EXPECT_EQ(axis->size(), converter.axisSize(0));
    EXPECT_EQ(axis->lowerBound(), converter.calculateMin(0, Axes::Coords::UNDEFINED));
    EXPECT_EQ(axis->upperBound(), converter.calculateMax(0, Axes::Coords::UNDEFINED));

    auto axis2 = converter.createConvertedAxis(1, Axes::Coords::RADIANS);
    EXPECT_TRUE(dynamic_cast<const FixedBinAxis*>(axis2.get()));
    EXPECT_EQ(axis2->size(), converter.axisSize(1));
    EXPECT_EQ(axis2->lowerBound(), converter.calculateMin(1, Axes::Coords::RADIANS));
    EXPECT_EQ(axis2->upperBound(), converter.calculateMax(1, Axes::Coords::RADIANS));

    EXPECT_THROW(converter.createConvertedAxis(2, Axes::Coords::UNDEFINED), std::runtime_error);
    EXPECT_THROW(converter.createConvertedAxis(1, Axes::Coords::QSPACE), std::runtime_error);
}

TEST_F(OffSpecularCoordinatesTest, OffSpecularCoordinatesClone)
{
    OffSpecularCoordinates converter(m_detector, m_beam, m_alpha_i_axis);
    std::unique_ptr<OffSpecularCoordinates> P_clone(converter.clone());

    EXPECT_EQ(P_clone->dimension(), 2u);

    EXPECT_DOUBLE_EQ(P_clone->calculateMin(0, Axes::Coords::UNDEFINED), 0.0);
    EXPECT_DOUBLE_EQ(P_clone->calculateMin(0, Axes::Coords::NBINS), 0.0);
    EXPECT_DOUBLE_EQ(P_clone->calculateMin(0, Axes::Coords::RADIANS), 0.0);
    EXPECT_DOUBLE_EQ(P_clone->calculateMin(0, Axes::Coords::DEGREES), 0.0);
    EXPECT_THROW(P_clone->calculateMin(0, Axes::Coords::QSPACE), std::runtime_error);
    EXPECT_THROW(P_clone->calculateMin(0, Axes::Coords::MM), std::runtime_error);

    EXPECT_DOUBLE_EQ(P_clone->calculateMax(0, Axes::Coords::UNDEFINED), 7.0);
    EXPECT_DOUBLE_EQ(P_clone->calculateMax(0, Axes::Coords::NBINS), 51.0);
    EXPECT_DOUBLE_EQ(P_clone->calculateMax(0, Axes::Coords::RADIANS), Units::deg2rad(7.0));
    EXPECT_DOUBLE_EQ(P_clone->calculateMax(0, Axes::Coords::DEGREES), 7.0);
    EXPECT_THROW(P_clone->calculateMax(0, Axes::Coords::QSPACE), std::runtime_error);
    EXPECT_THROW(P_clone->calculateMax(0, Axes::Coords::MM), std::runtime_error);

    EXPECT_DOUBLE_EQ(P_clone->calculateMin(1, Axes::Coords::UNDEFINED), -2.0);
    EXPECT_DOUBLE_EQ(P_clone->calculateMin(1, Axes::Coords::NBINS), 0.0);
    EXPECT_DOUBLE_EQ(P_clone->calculateMin(1, Axes::Coords::RADIANS), Units::deg2rad(-2.0));
    EXPECT_DOUBLE_EQ(P_clone->calculateMin(1, Axes::Coords::DEGREES), -2.0);
    EXPECT_THROW(P_clone->calculateMin(1, Axes::Coords::QSPACE), std::runtime_error);
    EXPECT_THROW(P_clone->calculateMin(1, Axes::Coords::MM), std::runtime_error);

    EXPECT_DOUBLE_EQ(P_clone->calculateMax(1, Axes::Coords::UNDEFINED), Units::rad2deg(1.5));
    EXPECT_DOUBLE_EQ(P_clone->calculateMax(1, Axes::Coords::NBINS), 70.0);
    EXPECT_DOUBLE_EQ(P_clone->calculateMax(1, Axes::Coords::RADIANS), 1.5);
    EXPECT_DOUBLE_EQ(P_clone->calculateMax(1, Axes::Coords::DEGREES), Units::rad2deg(1.5));
    EXPECT_THROW(P_clone->calculateMax(1, Axes::Coords::QSPACE), std::runtime_error);
    EXPECT_THROW(P_clone->calculateMax(1, Axes::Coords::MM), std::runtime_error);

    EXPECT_THROW(P_clone->calculateMin(2, Axes::Coords::UNDEFINED), std::runtime_error);
    EXPECT_THROW(P_clone->calculateMax(2, Axes::Coords::UNDEFINED), std::runtime_error);
}
